from datetime import timedelta

import numpy as np
import pytest

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.counters import Counter, CuttingBoard
from overcooked_simulator.game_items import Item, ItemInfo, ItemType
from overcooked_simulator.overcooked_environment import (
    Action,
    Environment,
    ActionType,
    InterActionData,
)
from overcooked_simulator.utils import create_init_env_time

layouts_folder = ROOT_DIR / "game_content" / "layouts"
environment_config_path = ROOT_DIR / "game_content" / "environment_config.yaml"
layout_path = ROOT_DIR / "game_content" / "layouts" / "basic.layout"
layout_empty_path = ROOT_DIR / "game_content" / "layouts" / "basic.layout"
item_info_path = ROOT_DIR / "game_content" / "item_info.yaml"

# TODO: TESTs are in absolute pixel coordinates still.


@pytest.fixture(autouse=True)
def test_file_availability():
    assert layouts_folder.is_dir(), "layouts folder does not exists"
    assert environment_config_path.is_file(), "environment config file does not exists"
    assert layout_path.is_file(), "layout config file does not exists"
    assert layout_empty_path.is_file(), "layout empty config file does not exists"
    assert item_info_path.is_file(), "item info config file does not exists"


@pytest.fixture
def env_config():
    with open(environment_config_path, "r") as file:
        env_config = file.read()

    return env_config


@pytest.fixture
def layout_config():
    with open(layout_path, "r") as file:
        layout = file.read()
    return layout


@pytest.fixture
def layout_empty_config():
    with open(layout_path, "r") as file:
        layout = file.read()
    return layout


@pytest.fixture
def item_info():
    with open(item_info_path, "r") as file:
        item_info = file.read()
    return item_info


def test_player_registration(env_config, layout_config, item_info):
    env = Environment(env_config, layout_config, item_info, as_files=False)

    env.add_player("1")
    assert len(env.players) == 1, "Wrong number of players"

    env.add_player("2")
    assert len(env.players) == 2, "Wrong number of players"

    env.add_player("2")
    assert len(env.players) == 2, "Wrong number of players"


def test_movement(env_config, layout_empty_config, item_info):
    env = Environment(env_config, layout_empty_config, item_info, as_files=False)
    player_name = "1"
    start_pos = np.array([3, 4])
    env.add_player(player_name, start_pos)
    env.players[player_name].player_speed_units_per_seconds = 1
    move_direction = np.array([1, 0])
    move_action = Action(player_name, ActionType.MOVEMENT, move_direction, duration=0.1)
    do_moves_number = 3
    for i in range(do_moves_number):
        env.perform_action(action=move_action)
        env.step(timedelta(seconds=0.1))

    expected = start_pos + do_moves_number * (
        move_direction
        * env.players[player_name].player_speed_units_per_seconds
        * move_action.duration
    )

    assert np.isclose(
        np.linalg.norm(expected - env.players[player_name].pos), 0
    ), "Performed movement do not move the player as expected."


def test_player_speed_units_per_seconds(env_config, layout_empty_config, item_info):
    env = Environment(env_config, layout_empty_config, item_info, as_files=False)
    player_name = "1"
    start_pos = np.array([3, 4])
    env.add_player(player_name, start_pos)
    env.players[player_name].player_speed_units_per_seconds = 2
    move_direction = np.array([1, 0])
    move_action = Action(player_name, ActionType.MOVEMENT, move_direction, duration=0.1)
    do_moves_number = 3
    for i in range(do_moves_number):
        env.perform_action(action=move_action)
        env.step(timedelta(seconds=0.1))

    expected = start_pos + do_moves_number * (
        move_direction
        * env.players[player_name].player_speed_units_per_seconds
        * move_action.duration
    )

    assert np.isclose(
        np.linalg.norm(expected - env.players[player_name].pos), 0
    ), "Performed movement do not move the player as expected."


def test_collision_detection(env_config, layout_config, item_info):
    env = Environment(env_config, layout_config, item_info, as_files=False)

    counter_pos = np.array([1, 2])
    counter = Counter(counter_pos)
    env.counters = [counter]
    env.add_player("1", np.array([1, 1]))
    env.add_player("2", np.array([1, 4]))

    player1 = env.players["1"]
    player2 = env.players["2"]

    assert not env.detect_collision_counters(player1), "Should not collide"
    assert not env.detect_player_collision(player1), "Should not collide yet."

    assert not env.detect_collision(player1), "Does not collide yet."

    player1.move_abs(counter_pos)
    assert env.detect_collision_counters(
        player1
    ), "Player and counter at same pos. Not detected."
    player2.move_abs(counter_pos)
    assert env.detect_player_collision(player1), "Players at same pos. Not detected."

    player1.move_abs(np.array([-1, -1]))
    assert env.detect_collision_world_bounds(
        player1
    ), "Player collides with world bounds."


def test_player_reach(env_config, layout_empty_config, item_info):
    env = Environment(env_config, layout_empty_config, item_info, as_files=False)

    counter_pos = np.array([2, 2])
    counter = Counter(counter_pos)
    env.counters = [counter]
    env.add_player("1", np.array([2, 4]))
    env.players["1"].player_speed_units_per_seconds = 1
    player = env.players["1"]
    assert not player.can_reach(counter), "Player is too far away."

    do_moves_number = 30
    for i in range(do_moves_number):
        move_action = Action("1", ActionType.MOVEMENT, np.array([0, -1]), duration=1)
        env.perform_action(move_action)
        env.step(passed_time=timedelta(seconds=1))
    assert player.can_reach(counter), "Player can reach counter?"


def test_pickup(env_config, layout_config, item_info):
    env = Environment(env_config, layout_config, item_info, as_files=False)

    counter_pos = np.array([2, 2])
    counter = Counter(counter_pos)
    counter.occupied_by = Item(name="Tomato", item_info=None)
    env.counters = [counter]

    env.add_player("1", np.array([2, 3]))
    player = env.players["1"]
    player.player_speed_units_per_seconds = 1

    move_down = Action("1", ActionType.MOVEMENT, np.array([0, -1]), duration=1)
    move_up = Action("1", ActionType.MOVEMENT, np.array([0, 1]), duration=1)
    pick = Action("1", ActionType.PUT, "pickup")

    env.perform_action(move_down)
    env.step(timedelta(seconds=1))
    assert player.can_reach(counter), "Player can reach counter?"

    env.perform_action(pick)
    assert player.holding is not None, "Player should have picked up tomato."
    assert player.holding.name == "Tomato", "Should be tomato."

    for _ in range(5):
        env.perform_action(move_up)
        env.step(timedelta(seconds=1))

    env.perform_action(pick)
    assert (
        player.holding is not None
    ), "Player should be too far away to put tomato down."

    for _ in range(4):
        env.perform_action(move_down)
        env.step(timedelta(seconds=1))
        env.perform_action(move_down)
        env.step(timedelta(seconds=1))
        env.perform_action(move_down)
        env.step(timedelta(seconds=1))
        env.perform_action(move_down)
        env.step(timedelta(seconds=1))

    env.perform_action(pick)

    assert player.holding is None, "Player should have put tomato down."
    assert (
        counter.occupied_by is not None and counter.occupied_by.name == "Tomato"
    ), "Tomato should be here now."


def test_processing(env_config, layout_config, item_info):
    env = Environment(env_config, layout_config, item_info, as_files=False)
    counter_pos = np.array([2, 2])
    counter = CuttingBoard(
        counter_pos,
        transitions={
            "ChoppedTomato": ItemInfo(
                name="ChoppedTomato",
                seconds=0.5,
                equipment=ItemInfo(name="CuttingBoard", type=ItemType.Equipment),
                type=ItemType.Ingredient,
                needs=["Tomato"],
            )
        },
    )
    env.counters.append(counter)
    env.progressing_counters.append(counter)

    tomato = Item(name="Tomato", item_info=None)
    env.add_player("1", np.array([2, 3]))
    player = env.players["1"]
    player.player_speed_units_per_seconds = 1
    player.holding = tomato

    move = Action("1", ActionType.MOVEMENT, np.array([0, -1]), duration=1)
    pick = Action("1", ActionType.PUT, "pickup")

    env.perform_action(move)
    env.step(timedelta(seconds=1))
    env.perform_action(pick)

    hold_down = Action("1", ActionType.INTERACT, InterActionData.START)
    env.perform_action(hold_down)

    assert tomato.name != "ChoppedTomato", "Tomato is not finished yet."

    env.step(timedelta(seconds=1))

    assert tomato.name == "ChoppedTomato", "Tomato should be finished."

    button_up = Action("1", ActionType.INTERACT, InterActionData.STOP)
    env.perform_action(button_up)
    env.perform_action(pick)
    assert player.holding.name == "ChoppedTomato", "Tomato should be finished."


def test_time_passed():
    np.random.seed(42)
    env = Environment(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        ROOT_DIR / "game_content" / "item_info.yaml",
    )
    env.reset_env_time()
    passed_time = timedelta(seconds=10)
    env.step(passed_time)
    assert (
        env.env_time == create_init_env_time() + passed_time
    ), "Env time needs to be updated via the step function"

    passed_time_2 = timedelta(seconds=12)
    env.step(passed_time_2)
    assert (
        env.env_time == create_init_env_time() + passed_time + passed_time_2
    ), "Env time needs to be updated via the step function"


def test_time_limit():
    np.random.seed(42)
    env = Environment(
        ROOT_DIR / "game_content" / "environment_config.yaml",
        layouts_folder / "empty.layout",
        ROOT_DIR / "game_content" / "item_info.yaml",
    )
    env.reset_env_time()

    assert not env.game_ended, "Game has not ended yet"

    passed_time = timedelta(seconds=10)
    env.step(passed_time)

    assert not env.game_ended, "Game has not ended yet"

    passed_time_2 = timedelta(
        seconds=(env.env_time_end - env.start_time).total_seconds()
    )
    env.step(passed_time_2)

    assert env.game_ended, "Game has ended now."
